@MirrorsUsed(targets: const [Expand], override: "*")
import 'dart:mirrors';
import 'dart:convert';

class Expand{
	const Expand();
}

class Exportable {
	static bool _isExportableClassValue(Object value) {
		return value is Exportable;
	}

	static Type _getActualVariableType(VariableMirror variableMirror) {
		if (variableMirror.type.qualifiedName == reflectType(List).qualifiedName) {
			return variableMirror.type.typeArguments[0].reflectedType;
		}
		return variableMirror.type.reflectedType;
	}

	static bool _isJsonSupported(value) {
		if (value == null || value is bool || value is num || value is String) {
			return true;
		}
		return false;
	}

	static dynamic _importSimpleValue(Type type, value) {
		if (type == DateTime && value is String) {
			return DateTime.parse(value).toLocal();
		} else if (_isJsonSupported(value)) {
			return value;
		}
		return null;
	}

	static dynamic _exportSimpleValue(value) {
		if (_isJsonSupported(value)) {
			return value;
		} else if (value is DateTime) {
			return (value as DateTime).toUtc().toString();
		}
		return null;
	}

	static Object getInstatiatedTypeForJson(Type type, json, [bool areEqual(String key, String variableName) = null]) {
		if (areEqual == null) areEqual = (key, name) => key == name;

		if (json is String) json = JSON.decode(json);
		if (json == null) return null;
		if (json is List) {
			var result = [];
			json.forEach((e) => result.add(getInstatiatedTypeForJson(type, e, areEqual)));
			return result;
		} else {

			var result = reflectClass(type).newInstance(const Symbol(''), []).reflectee;
			InstanceMirror thisMirror = reflect(result);

			var delerationMirrors = thisMirror.type.declarations.values.toList();

			for (var type = thisMirror.type.superclass; type.reflectedType != Object; type = type.superclass) {
				delerationMirrors.addAll(type.declarations.values.toList());
			}

			for (DeclarationMirror declarationMirror in delerationMirrors) {
				if (declarationMirror is VariableMirror) {
					var name = MirrorSystem.getName(declarationMirror.simpleName);
					var actualVariableType = _getActualVariableType(declarationMirror);
					var matchedKeys = json.keys.where((String key) => areEqual(key, name));

					if (matchedKeys.length > 0) {
						var value = json[matchedKeys.elementAt(0)];
						if (!_isJsonSupported(value)) {
							thisMirror.setField(declarationMirror.simpleName, getInstatiatedTypeForJson(actualVariableType, value, areEqual));
						} else {
							thisMirror.setField(declarationMirror.simpleName, _importSimpleValue(declarationMirror.type.reflectedType, value));
						}
					}
				}
			}
			return result;
		}
	}

	static String getJsonFromObject(Object object, [String getName(String name) = null]) {
		if (getName == null) getName = (name) => name;
		return JSON.encode(_getJsonFromObject(object, getName));
	}

	static Object _getJsonFromObject(Object object, String getName(String name)) {
		if (object is List) {
			var list = [];
			for (var obj in object) {
				list.add(_getJsonFromObject(obj, getName));
			}
			return list;
		} else {
			var mapResult = {};
			InstanceMirror thisMirror = reflect(object);
			var delerationMirrors = thisMirror.type.declarations.values.toList();

			for (var type = thisMirror.type.superclass; type.reflectedType != Object; type = type.superclass) {
				delerationMirrors.addAll(type.declarations.values.toList());
			}
			//thisMirror.type.superclass.reflectedType == Exportable
			for (DeclarationMirror declarationMirror in delerationMirrors) {
				if (declarationMirror is VariableMirror) {
					var name = getName(MirrorSystem.getName(declarationMirror.simpleName));
					var value = thisMirror.getField(declarationMirror.simpleName).reflectee;
					if (value is List) {

						mapResult[name] = value.map((val) => _isExportableClassValue(val) ? _getJsonFromObject(val, getName) : _exportSimpleValue(val)).toList();
					} else {
						mapResult[name] = _isExportableClassValue(value) ? _getJsonFromObject(value, getName) : _exportSimpleValue(value);
					}
				}
			}
			return mapResult;
		}
	}
}
